/****************************************************************************/
/*                                                                          */
/*  File name: Ptcp6.c                                                       */
/*  Purpose  : Send via TCP a buffer to a SERVER(mirror) and wait for an    */
/*             echo back. The sender is the client                          */
/*                                                                          */
/*  Author   : J.ROCHEZ                                                     */
/*             Basic code copied mainly from ttcp.c                         */
/*                                                                          */
/*  CERN, Geneva 1997 - Copyright and any other appropriate                 */
/*  legal protection of these computer programs and associated              */
/*  documentation reserved in all countries of the world.                   */
/*                                                                          */
/* ------------------------------------------------------------------------ */
/*  Version  |  Author  |  Mod.Date  |            Changes                   */
/* ------------------------------------------------------------------------ */
/*    1      |    JR    | 31-JUL-1997| Original coding,                     */
/*           |          |            | for SYSV and WNT (sun,lynx,WNT)      */
/*           |          |            | NO coding for BSD4.x                 */
/****************************************************************************/
/*              system dependent setting
 *              ------------------------
 * uname -a,gcc -v a.c are the tools used
 *
 * Platform/OS    #define         MACRO predefined
 * -----------    -------   ---------------------------------------------------
 * SunOS OS       BSD43     __unix__   __sun__   __sparc__
 * SUN Solaris     SYSV     __unix__   __sun__   __sparc__   __svr4__
 * SGI-IRIX  < 3.3 SYSV     set as #define sgi
 * HP-UX 9000      SYSV     __unix__   __hpux__  __hp9k8__
 * OSF/1 V3.2      SYSV     __unix__   __osf__   __alpha__
 * OSF/1 V4.0      SYSV     __unix__   __osf__   __alpha__   _CFE
 * LynxOS          SYSV     __unix__   __lynx__  __powerpc__
 * Windows NT      SYSV                __WINNT__ __i386__    __M_COFF__
 * AIX             SYSV     _AIX       _AIX32    _POWER      _IBMR2   _ARCH_PWR
 * Micosoft Visual C++ compiler under WindowNT 4.0
 * Windows NT               _WINDOWS    WIN32
 
 *
 *            Compiler commands
 *            -----------------
 * LynxOS : gcc -c Ptcp6.c -o Ptcp6.o | gcc -o Ptcp6 -O Ptcp6.o -lnetinet -lc_p -lc
 */
 
/* -------------attempt to set an automatic UNIX  OS family detection -------*/


#define SYSV
 
#if defined(__hp9k8__) || defined(__osf__) || defined(__srv4__) 
#define SYSV
#endif
#if defined (__lynx__)
#define SYSV
#endif
 
/* version A.09.01  'key words' differs from A.09.05 A */
#if defined(__hpux)
#define __hpux__
#endif

#if defined(_WIN32)        /* Window NT 4.0 compiled with VisualC++ 4 */
#define __NTVIS__
#define SYSV
#endif

/*--------------------------------------------------------------------------*/
#include <stdio.h>
#include <signal.h> 
#include <ctype.h>
#include <errno.h>
#include <sys/types.h>

/* all systems are SYSV   */
 
#if defined(__osf__)
#include <sys/types.h>     
#include <sys/socket.h>         
#include <netinet/in.h>
/* Start INET6 */
#include <netinet/in6.h>
/* End INET6 */
#include <netinet/tcp.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <sys/time.h>      /* struct timeval */
 
#else  /* else of __osf__ */

#if defined(__NTVIS__)
#include <windows.h>       /* required for all Windows applications */
/* #include <winsock.h>      */
#include <memory.h>
#include <time.h>
/* #include <sys\types.h>     bug in Visual C about def of time_t. To clear it,  
                            * sys/types.h must be defined after time.h */
#include <sys\timeb.h>
#include <process.h>       /* for _beginthread                      */
#include <stdlib.h>
 
#define bcopy(a,b,n) memcpy((b), (a), (n))
#define bzero(a,n) memset((a), 0, (n))
extern int getopt();
 
#else  /* else of __NTVIS__ */

#if defined(__lynx__)
#include <sys/types.h>     
#include <socket.h>        /* located in  /usr/include/..... */
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <time.h>
#include <sys/times.h>

#include <conf.h>          /* definition of TICKSPERSEC (HZ) */
#include <sys/param.h>
 
#else  /* else of __Lynx__  */
#include <sys/types.h>     
#if defined (__svr4__)
#include <sys/socket.h>
#include <sys/uio.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <sys/time.h>      /* struct timeval */
#include <sys/times.h>
#include <sys/param.h>
 
#else  /* else of __svr4__    all SYSV cases except those mentionned before */
#include <sys/types.h>     
#include <sys/socket.h>
#include <netinet/in.h>

#include <netinet/tcp.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <sys/time.h>      /* struct timeval */
#include <sys/times.h>
#include <sys/param.h>
 
#endif /* __svr4__  */
#endif /* __lynx__  */
#endif /* __NTVIS__ */
#endif /* __osf__   */


/* structures --------------------------------------------------------------*/
#if defined(__sun__)
/* Start INET6 */
 struct sockaddr_in6 sockaddr;   /* done in ./X11/Xdmcp.h */
#endif
 
 struct sockaddr_in6 sa_him;   /* socket struct. set with local parameters */
 struct sockaddr_in6 sa_me;   /* socket struct. image of the remote host */

 
#if defined(__lynx__) || defined(__svr4__) || defined(_AIX)
 struct sockaddr_in6 frominet;
#else
 struct sockaddr_in6 frominet;
#endif /* __lynx__ */
/* End INET6 */
 
 struct timeval tstart,tend;               /* start and end */ 
 struct tm *tms;                           /* system time structure */

 struct hostent *addr;

/* extern declarations -----------------------------------------------------*/
#if !defined (__lynx__)
 extern int errno;
#endif

#if defined(__NTVIS__)
/* unix like commands under WindowsNT -------------------------------------*/
#ifndef _GETOPT_
#define _GETOPT_

int getopt(int argc, char **argv, char *optstring);

extern char *optarg;        // returned arg to go with this option
extern int optind;        // index to next argv element to process
extern int opterr;        // should error messages be printed?
extern int optopt;        //

#define BADCH ('?')

#endif // _GETOPT


/* get option letter from argument vector   */
int
    opterr = 1,        // should error messages be printed?
    optind = 1,        // index into parent argv vector
    optopt;            // character checked for validity
char
    *optarg;           // argument associated with option

#define EMSG    ""

char *progname;            // may also be defined elsewhere

#endif /* __NTVIS__ */

extern int optind;
extern char *optarg;




/* socket declaration ------------------------------------------------------*/
#if defined(__NTVIS__)
 SOCKET fd;                  /* network socket field descriptor */
 SOCKET nfd;
#else
 int fd,nfd;                     /* nfd == socket created by accept */
#endif /* __NTVIS__ */


 /* variables definitions */

 

 int    sockbufsize = 0;         /* socket buffer size to use */
 FILE   *fplog;                  /* file pointer for the log file */
 char   logfile[100];            /* file name for the log */
static char logfile_head[] = "ptcp6_log";  /* header name of the log */
 int    c;
 int    client;                  /* 0=server,  !0=client */
 int    forever = 0;             /* loop for ever */
 int    verbose = 0;             /* 0=print basic info  */
 int    nodelay = 0;             /* set TCP_NODELAY socket option */
 int    one = 1;                 /* for 4.3 BSD style setsockopt() */

 char   *host;                   /* ptr to name of host */
 struct hostent *addr;           /* to fix the ip address */ 
 unsigned long addr_tmp;         /* temp address */
 short  port = 5010;             /* TCP port number */

 int    buflen = 8 * 1024;       /* length of transmit buffer */
 char   *txbuf;                  /* ptr to dynamic transmit buffer */
 char   *rxbuf;                  /* ptr to dynamic received buffer */
 int    nbuf = 2 * 1024;         /* number of buffers to send in sinkmode */
 int    bufoffset = 0;           /* align buffer (tx and rx) to this */
 int    bufalign = 16*1024;      /* modulo this */

 int    log_cnt = 0;             /* append result to a log */
 int    validity = 1;            /* == 0,  if time rounded to t_min */
 int    debug = 0;

 # define MAX_STRING 132
 char stats[MAX_STRING];
 double tx_bytes;                /* bytes sent on net */
 double rx_bytes;                /* bytes received from the net */
 double rate;                    /* average tx rate */
 unsigned long numCalls;         /* # of I/O system calls */
 double delay_t;                 /* real time (useconds) of send and receive */
 double tx_t;                    /* time to sent all buffers */

/*-------------------------------------------------------------------------*/
/*-----------Prototype functions definitions -------------------------------*/
/*--------------------------------------------------------------------------*/
 
                                   /* ANSI input/output functions (stdio.h) */
 
#if defined (__lynx__)
int  getopt(int, char**, char*);
int  gettimeofday(struct timeval *tp, struct timezone *tzp);
 
#else
#if defined (__svr4__)
 
#else
#if defined (_AIX)
 
#else
#if defined (__hpux__)
#else
#if defined (__NTVIS__)
int  gettimeofday(struct timeval *tp, struct timezone *tzp); 

#else

#endif  /* __hpux__ */
#endif  /*  _AIX    */
#endif  /* __svr4__ */
#endif  /* __lynx__ */
#endif  /* __NTVIS__ */


void main(int argc, char* argv[]);
int  read(int, char*, int);
int  write(int, char*, int);
int  close(int);
int fclose(FILE *stream);
 
#if !defined (__NTVIS__)
int  atoi(char*);
char strncpy(char *s1,char *s2,size_t n);
void bzero(char*,int);
void bcopy(char*, char*, int);
int  malloc(int);
#endif
                          /* ANSI socket functions prototype /sys/socket.h */
#if defined (__lynx__)
int  select(int, fd_set*, fd_set*, fd_set*, struct timeval*);
 
#else
#if defined (__svr4__)

#else
#if defined (_AIX)
int  select(unsigned long, void *, void *, void *, struct timeval *);
 
#else
#if defined(__hpux__)
 
#else
#if defined (__NTVIS__)
 
 

#endif /* __hpux__ */
#endif /*  _AIX    */
#endif /* __svr4__ */
#endif /* __lynx__ */
#endif /* __NTVIS__ */

/* Ptcp6 prototype functions */

int    err(char *);
void   mes(char *);
void   do_usage(void);
void   do_help(void);
int    do_write(int,void*,int);
int    do_read(int,void*,int);
void   prep_timer(void);
void   open_log(void);
void   close_log(void);
void   get_options(int argc, char *argv[]);
double delta_time(struct timeval *, struct timeval *); 
void   close_task(void);

#if !defined(__NTVIS__)
/*
 void sigpipe(void);
 lynx void(*signal (sig, func))()
      int sig;
      void(*func)();
      void (*signal (int, void (*)(int))) (int); ... rchrch c'est du chinois
*/
#endif
 


/*--------------------------------------------------------------------------*/
#if !defined(__NTVIS__)
void
sigpipe()
{;
}
#endif
/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/
void
main(argc, argv)
int argc;
char *argv[];
{ 
 int i;
 char buffer[51];
 int domain,fromlen;
 double round,latency;
 int sockbufsndsize,sockbufrcvsize;
 int sockbuflen;
/* Start INET6 */
 struct sockaddr_in6 peer;
/* End INET6 */
 int peerlen = sizeof(peer);
 register int nb,txcnt,rxcnt,rxcnt_total,multi;


#if defined (__NTVIS__)
 extern char *optarg;
 WSADATA WSAData;
 WSAStartup(MAKEWORD(1,1), &WSAData);
 /* withou this we have an execution error 10093 (socket assignment error ) */

#endif /* __NTVIS__ */






 /* collect all input options -----------------------------------------------*/
 get_options(argc,argv);

 if (log_cnt) open_log();

 /* reserve buffers space ---------------------------------------------------*/
 if ( (txbuf = (char *)malloc(buflen+bufalign)) == (char *)NULL)
      err("malloc");
 if (bufalign != 0)
      txbuf +=(bufalign - ((int)txbuf % bufalign) + bufoffset) % bufalign;

 if ( (rxbuf = (char *)malloc(buflen+bufalign)) == (char *)NULL)
      err("malloc");
 if (bufalign != 0)
      rxbuf +=(bufalign - ((int)rxbuf % bufalign) + bufoffset) % bufalign;
 
 fprintf(stdout,"Ptcp6%s: tx& rx buffer buflen=%d, nbuf=%d, align=%d/%d, \n",
            client?"-c":"-s",buflen, nbuf, bufalign, bufoffset);
 if (log_cnt)fprintf(fplog," %6d %6d %6d %6d %4d",
                      buflen, nbuf, bufalign, bufoffset, port);

 /* preload the buffer ------------------------------------------------------*/

 if (client)
   {i = buflen;
    /* while (i) {txbuf[i-1] = (char)((i-1) % 256);  --i; }  fill ascii */
    /* other way do do it 
     * register char c; register int i;
     * while (i-- > 0) 
     *    { while ( !isprint(c&0x7F)) ) c++;  *buf++ = (c++&0x7F);} 
     */
   } 
/* set the socket address input and output --------------------------------- */

 
 if (client)  /* the one sending first a buffer */
   {
   /* get the destination ip number or host name */
   if (optind == argc) 
     {fprintf(stdout,"Ptcp6%s: Missing host name.\n\n",client?"-c":"-s");
       do_usage(); exit(1); }
   bzero((char *)&sa_him, sizeof(sa_him));
   host = argv[optind];
   if (atoi(host) > 0 )  
     {/* Numeric */
/* Start INET6 */
      sa_him.sin6_family = AF_INET6;
      inet_pton(AF_INET6,host,sa_him.sin6_addr.s6_addr);
     } 
   else 
     {if ((addr=gethostbyname2(host, AF_INET6)) == NULL) err("bad hostname");
      sa_him.sin6_family = addr->h_addrtype;
      sa_him.sin6_family = addr->h_addrtype;
      bcopy(addr->h_addr,(char*)&(sa_him.sin6_addr.s6_addr), addr->h_length);
     }
   sa_him.sin6_port = htons(port);
   sa_me.sin6_port = 0;
   }
 else  
   {
    sa_me.sin6_port = htons(port);
   }
 
#if defined(__NTVIS__)
 sa_me.sin6_family = AF_INET6;
#endif
/* End INET6 */

if (verbose)
  { if (client)   
      fprintf(stdout,"Ptcp6%s: tcp client(sender) -> %s on port %d.\n",
                      client?"-c":"-s",host,port);
    else fprintf(stdout,"Ptcp6%s: tcp server(mirror) on port %d.\n",
                      client?"-c":"-s",port);
  }

 /* assign socket ----------------------------------------------------------*/
                             
#if defined(__NTVIS__)
 if ((fd = socket(AF_INET6, SOCK_STREAM, 0)) == INVALID_SOCKET) 
 { err("socket");}
#else
 if ((fd = socket(AF_INET6,SOCK_STREAM, 0)) < 0) err("socket");
#endif
 
 if (verbose) 
    fprintf(stdout,"Ptcp6%s: File-Descriptor  0x%x Opened\n" ,
                      client?"-c":"-s", fd );
               



/* set socket options -----------------------------------------------------*/
/* (1) Socket size for the server and client (if possible) ----------------*/

#if defined(SO_SNDBUF) || defined(SO_RCVBUF)
 if (sockbufsize)
   {
#if defined(__lynx__) || defined(__sun__)  || defined (__NTVIS__)
    if (setsockopt(fd, SOL_SOCKET, SO_SNDBUF, (char *)&sockbufsize,
                       sizeof sockbufsize) < 0)  err("setsockopt: sndbuf");
    if (setsockopt(fd, SOL_SOCKET, SO_RCVBUF, (char *)&sockbufsize,
                       sizeof sockbufsize) < 0)  err("setsockopt: rcvbuf");
#else 
    if (setsockopt(fd, SOL_SOCKET, SO_SNDBUF, &sockbufsize,
                       sizeof sockbufsize) < 0)  err("setsockopt: sndbuf");
    if (setsockopt(fd, SOL_SOCKET, SO_RCVBUF, &sockbufsize,
                       sizeof sockbufsize) < 0)  err("setsockopt: rcvbuf");
#endif /* __lynx__ __sun__ __NTVIS__ */
   }
 else
   {/* the socket buffer size is set by the system to read it : 
     * 
    ** Added by Daniel Davids to Know Socket-Buffer-Sizes
    */
     sockbuflen = sizeof sockbufsndsize;
   #if defined(__lynx__) || defined(__sun__) || defined (__NTVIS__)
     getsockopt(fd, SOL_SOCKET, SO_SNDBUF, (char *)&sockbufsndsize,
                &sockbuflen);
     sockbuflen = sizeof sockbufrcvsize;
     getsockopt(fd, SOL_SOCKET, SO_RCVBUF, (char *)&sockbufrcvsize,
                     &sockbuflen);
   #else
     getsockopt(fd, SOL_SOCKET, SO_SNDBUF, &sockbufsndsize, &sockbuflen);
     sockbuflen = sizeof sockbufrcvsize;
     getsockopt(fd, SOL_SOCKET, SO_RCVBUF, &sockbufrcvsize, &sockbuflen);
   #endif /* __lynx__ __sun__ __NTVIS__ */

     sockbufsize = ( sockbufsndsize + sockbufrcvsize ) / 2;
     if ( sockbufsndsize != sockbufrcvsize )
       {fprintf(stdout, "sockbufSNDSIZE=%d, ", sockbufsndsize );
        fprintf(stdout, "sockbufRCVSIZE=%d, ", sockbufrcvsize );
       }
   }
  #endif /* defined(SO_SNDBUF) || defined(SO_RCVBUF) */
 
 if (sockbufsize) fprintf(stdout, "sockbufsize=%d, \n", sockbufsize);
 
 if (log_cnt)
   {if (sockbufsize)fprintf(fplog," %6d",sockbufsize);
    else fprintf(fplog," 0");
   }

/* (2) set the NODELAY option------------------------------------------------*/ 
#ifdef TCP_NODELAY
 if (nodelay)
   {struct protoent *p;
    p = getprotobyname("tcp");
  #if defined (__lynx__)  || defined(__sun__) || defined (__NTVIS__)
    if( p && setsockopt(fd, p->p_proto, TCP_NODELAY,
              (char *) &one, sizeof(one)) < 0)
  #else
    if( p && setsockopt(fd, p->p_proto, TCP_NODELAY,
              &one, sizeof(one)) < 0)
  #endif /* __lynx__  __sun__ __NTVIS__ */
    err("setsockopt: nodelay");
   }
#endif /* TCP_NODELAY */


 /* signal ??? --------------------------------------------------------------*/
#if !defined(__NTVIS__)
 signal(SIGPIPE, sigpipe);
#endif /* !__NTVIS__ */

 /* bind the socket with the socket address image ---------------------------*/
#if defined(__NTVIS__)
   if (bind(fd, (struct sockaddr FAR *)&sa_me,sizeof(sa_me))==SOCKET_ERROR)
     {fprintf(stderr, "bind: %d is the error\n", WSAGetLastError());
      perror("bind"); exit(1);
     }
#else
#if defined(__lynx__) || defined (__sun__) || defined (_AIX)
   if (bind(fd,(struct sockaddr *) &sa_me, sizeof(sa_me)) <0 ) err("bind");
#else
   if (bind(fd, &sa_me, sizeof(sa_me)) < 0)   err("bind");
#endif /* __sun__  __lynx__ _AIX */
#endif /* __NTVIS__ */

 /* connect the fd socket to the peer one ---------------------------------*/

if (client)
  {
 #if defined(__lynx__) ||defined(__svr4__) ||defined(_AIX) ||defined(__NTVIS__)
   if (connect(fd, (struct sockaddr *)&sa_him, sizeof(sa_him) ) < 0)
  #else
   if(connect(fd, &sa_him, sizeof(sa_him) ) < 0)
  #endif /* __lynx__ __svr4__ _AIX __NTVIS__ */
     err("connect");
  }
 else 
  {/* we are the server and we listen for a connection */
   errno = 0;
  #if defined(sgi) || ( defined(__osf__) && ! defined(_CFE) )
   if( listen(fd,1) <0 )   /* workaround for alleged u4.2 bug */
  #else
   if( listen(fd,0) <0 )   /* allow a queue of 0 */
  #endif
     err("listen");
   fromlen = sizeof(frominet);
/* Start INET6 */
   domain = AF_INET6;
/* End INET6 */

 /* accept the socket connection --------------------------------------------*/
 
  #if defined (__NTVIS__)
   if((nfd = accept(fd , (struct sockaddr * ) &frominet, &fromlen) ) < 0)
                err("accept");
  #else
   if((nfd = accept(fd, &frominet, &fromlen) ) < 0) err("accept");
  #endif /* __NTVIS__ */

  #if defined(__lynx__) ||defined(__sun__) ||defined(_AIX) ||defined(__NTVIS__)
   if (getpeername(nfd,(struct sockaddr *) &peer, &peerlen) < 0)
                   err("getpeername");
  #else
   if (getpeername(nfd,(struct sockaddr_in *) &peer,&peerlen) < 0)
                   err("getpeername");
   #endif /* __lynx__ __sun__ _AIX __NTVIS__ */
 
/* Start INET6 */
        fprintf(stdout,"Ptcp6-s: accept from %s\n", inet_ntop(AF_INET6,&(peer.sin6_addr),buffer,50));
/* End INET6 */
   } /* we are ... */

/* we do work (sending_receive  or mirroring )-------------------------------*/
 
 loop_entry:
 gettimeofday(&tstart, (struct timezone *)0);    /* start time */
 errno = 0;
 tx_bytes = 0.0;
 rx_bytes = 0.0;
 nb = 0;
 multi = nbuf;

 if (client)
   {
    while(multi--)
      {/* we send a buffer */
       txbuf[0] = multi;                      /* set flag in the buffer */

          
       txcnt=do_write(fd,txbuf,buflen);
       if (txcnt != buflen) 
         {fprintf(stdout,"ttcp%s: failure in the sending command.\n",
                 client?"-c":"-s");
          close_task(); exit(1);
         }  

       if(debug)fprintf(stdout,"Ptcp6%s: %5d | %d Bytes Written\n",
                         client?"-c":"-s", ++nb, txcnt );
       tx_bytes += buflen;

       /* now we wait and read a buffer 'buflen' back */
       rxcnt_total = 0; i =0;
       while (rxcnt_total < buflen)
         {rxcnt=do_read(fd,rxbuf,buflen);
          if (rxcnt < 0)
            {fprintf(stdout,"ttcp%s: failure in the receiving command.\n",
                     client?"-c":"-s");
             close_task(); exit(1);
            }
          if ( ( i ==0 ) && (txbuf[0] != rxbuf[0]) )
            {fprintf(stdout, "Ptcp6%s: Byte[0] in Tx and Rx buffer differ.\n",
                               client?"-c":"-s");
             close_task(); exit(1);
            }
          i++;
          if(debug)fprintf(stdout,"Ptcp6%s: %5d | %d Bytes received \n",
                           client?"-c":"-s", nb, rxcnt );
          rxcnt_total += rxcnt;
         }
       rx_bytes += rxcnt_total;
       if (rxcnt_total != txcnt)
         {fprintf(stdout, "Ptcp6%s: Qty of bytes sent", client?"-c":"-s");
          fprintf(stdout," %d and received %d differs\n",txcnt,rxcnt_total);
          close_task(); exit(1);
         }
      }
   /* sending,receiving  of nbuf time the buffer buf done  -----=------------*/
   }
 else 
   {/* we are server read the buffer and sent it back */
    multi = 0;
    while ((rxcnt=do_read(nfd,rxbuf,buflen)) > 0)
       {
        if (rxcnt < 0) 
         {fprintf(stdout,"ttcp%s: failure in the receiving command.\n",
              client?"-c":"-s");
              close_task(); exit(1);
             }
    if(debug)fprintf(stdout,"Ptcp6%s: %5d | %d Bytes Read( rxcnt) \n",
                             client?"-c":"-s", multi, rxcnt );
    rx_bytes += (double)rxcnt;

    /* we sent it back */
        if ( (txcnt=do_write(nfd,rxbuf,rxcnt)) != rxcnt) 
           {fprintf(stdout,"ttcp%s: failure in the sending command.\n",
                    client?"-c":"-s");

            close_task();  
           }

        if(debug)fprintf(stdout,"Ptcp6%s: %5d | %d Bytes Written\n",
                         client?"-c":"-s", multi, txcnt );
         tx_bytes += (double)txcnt;
         ++multi;
    }
   }
 /* done */
 gettimeofday(&tend, (struct timezone *)0);     /* End time */

 if(errno) err("IO");

 if(forever)
   {rx_bytes = 0; tx_bytes = 0;
    goto loop_entry;
   } 
 
 
 if (client)
   {
    delay_t = delta_time(&tend,&tstart);           /* send and receive delay */

   /* average bandwith */
   round = delay_t / (double)nbuf;              /* average time per buffer */ 
   latency = round / 2.0;                /* time for a buffer to be received */
   rate = ( (double)(buflen*nbuf)* 2.0) /(delay_t*1.048576);
   }
 else 
   {
   round = 0.0;
   latency = 0.0;
   rate = 0.0;
   validity = 1;
   }

 if (client)
   {
    
    fprintf(stdout,"Ptcp6%s: Buffer of %d bytes sent %d times\n",
               client?"-c":"-s",buflen, nbuf);
    fprintf(stdout,"Ptcp6%s: In total round trip delay of %.0f usec \n",
        client?"-c":"-s",delay_t);
    
    fprintf(stdout,"Ptcp6%s: Average transfer rate: %.3f Mbytes/sec\n",
        client?"-c":"-s", rate);
    fprintf(stdout,"Ptcp6%s: Average round trip   : %6.0f usec per buffer.\n",
           client?"-c":"-s",round);
    fprintf(stdout,"Ptcp6%s: Average latency delay: %6.0f usec per buffer.\n",
           client?"-c":"-s",latency);
   }
else  
   {
    fprintf(stdout,"Ptcp6%s: %.0f bytes echoed", client?"-c":"-s",rx_bytes);
    fprintf(stdout," in an average buff size of %.0f bytes\n", rx_bytes/(double)multi);
   }
if (log_cnt)
  {
   fprintf(fplog," %d %6.0f %6.0f %6.3f",validity,round,latency,rate); 
  }

close_task();

 
 if (verbose) fprintf(stdout,"Ptcp6%s: File-Descriptor  fd 0x%x Closed\n" ,
                     client?"-t":"-r", fd );
 
 fprintf(stdout,"Ptcp6 done.\n");
 exit(0);
 
}
/*-------------------------------------------------------------------------*/
void
close_task()
{
#if defined (__NTVIS__)
 closesocket ( fd );
 closesocket ( nfd );
#else
 close ( fd );
 close ( nfd );
#endif /* __NTVIS__ */
 
 if (log_cnt)    {fflush(stdout);  close_log(); }
}
/*--------------------------------------------------------------------------*/
void
close_log()
{

 static long sysTicks;
 time(&sysTicks);
 tms = localtime(&sysTicks);
     fprintf(fplog," %d\n",
             ((tms->tm_hour)*3600 + (tms->tm_min)*60 + tms->tm_sec));
 fclose(fplog);
 fflush(fplog);
}
/*--------------------------------------------------------------------------*/
double 
delta_time(t1,t0)
struct timeval *t1,*t0;
{
  double t;
  double t_min;
  struct timeval tdiff;

  tdiff.tv_sec = t1->tv_sec - t0->tv_sec;
  tdiff.tv_usec = t1->tv_usec - t0->tv_usec;
  if (tdiff.tv_usec < 0)  tdiff.tv_sec--, tdiff.tv_usec += (time_t)1000000;

  t = (double)tdiff.tv_sec * 1000000.0 + (double)tdiff.tv_usec;
 if (t <= 0.0)
   {fprintf(stdout,
    "Ptcp6%s: send delay time negatif, or value too big \n",  client?"-c":"-s");
    close_task(); exit(1);
   }
 t_min = 10.0;

 if( t <= t_min)              /* value in usec */
   { t = t_min;
    fprintf(stdout,"Ptcp6%s: send delay too short, set at ",client?"-c":"-s");
    fprintf(stdout,"%.0f usec, NOT accurate result.\n",t_min);
    validity = 0;    
   }
 else validity = 1;
 return(t);
}
/*--------------------------------------------------------------------------*/
/* write into the socket ----------------------------------------------------*/

int
do_write( s, bufp, count )
int s;
void *bufp;
int count;
{
  register int cnt;
#if defined (__NTVIS__)
/*   struct timeval tv;
  select(1,0,s,0,&tv);
  while (tv.tv_usec >= (time_t)0 )
  {printf(" aaa we wait \n"); }
 */
  cnt = send( s, bufp, count, 0 ); 
  numCalls++;
 #else
  cnt = write( s, bufp, count );
  numCalls++;
 #endif /* __NTVIS__ */
 
 return(cnt);
}

/* read the socket ---------------------------------------------------------*/

int
do_read( s, bufp, count )
int s;
void *bufp;
int count;
{
register int cnt;
 /*  rchrch option B in ttcp.c ???
   if( b_flag )  cnt = mread( s, bufp, count );         fill bufp 
  */

#if defined(__NTVIS__)
  cnt = recv( s, bufp, count, 0);
#else
  cnt = read( s, bufp, count );
#endif /* __NTVIS__ */
  numCalls++;

  return(cnt);
}


/*--------------------------------------------------------------------------*/
int
err(s)
char *s;
{fprintf(stderr,"Ptcp6%s: ", client?"-c":"-s");
 perror(s);
 fprintf(stderr,"Ptcp6%s: errno=%d\n",client?"-c":"-c",errno);
 close_task();
 exit(1);
 return(0);
}
/*--------------------------------------------------------------------------*/
void
get_options(argc,argv)
int argc;
char *argv[];

{
  if (argc < 2) { do_usage();     exit(1);}
  while ((c = getopt(argc, argv, "xFLDhcsvb:n:O:A:l:p:")) != -1) 
    {
      switch (c) 
       {
    case 'h':
      do_help();
      exit(1);
      break;
        case 'x':
                debug = 1;
                break;
    case 'c':
      client = 1;
      break;
    case 'F':
      forever = 1;
      break;
    case 's':
      client = 0;
      break;
    case 'v':
      verbose = 1;
      break;
        case 'D':
#ifdef TCP_NODELAY
        nodelay = 1;
#else
        fprintf(stderr,
       "Ptcp6: -D option ignored: TCP_NODELAY socket option not supported\n");
#endif
          break;
       case 'L':
      log_cnt = 1;
      break;
       case 'b':
#if defined(SO_SNDBUF) || defined(SO_RCVBUF)
          sockbufsize = atoi(optarg);
#else
          fprintf(stderr,
"Ptcp6: -b option ignored: SO_SNDBUF/SO_RCVBUF socket options not supported\n");
#endif
          break;
    case 'l':
      buflen = atoi(optarg);
      break;
        case 'A':
      bufalign = atoi(optarg);
          break;
       case 'O':
          bufoffset = atoi(optarg);
      break;
       case 'n':
          nbuf = atoi(optarg);
      break;
    case 'p':
      port = atoi(optarg);
      break;
    default:
      {do_usage(); exit(1);}
       }/*switch */
    }/* while */
 fprintf(stdout,"Ptcp6%s: started in mode : %s \n",
     client?"-c":"-s",client?"CLIENT":"SERVER");
}
/*--------------------------------------------------------------------------*/
void
do_help()
{
 char More_help[] = "\n\n\
               Client                      Server\n\
  Example  : ptcp6 -c host                 ptcp6-s \n\n\
  Total round trip  : is the time for a buffer of 'l' bytes to be \n\
                      sent to the server and received back repeated 'n' times.\n\
  Rate              : is the quantity of bytes passed on the network divided\n\
                      by the total round trip\n\
  Round trip        : is the total round trip divided by 'n'\n\
  latency           : is the round trip divided by 2. It is the average time\n\
                      for a frame to be writed in the memory of the server.\n\
";
 fprintf(stdout,More_help);
}

void
do_usage()
{
 char Usage[] = "\
  Usage: ptcp6 -c [-options] host        ptcp6 -s [-options] \n\n\
Common options:\n\
    -h      more help\n\
    -l ##   length of bufs read or written on the network (default 8192)\n\
    -A ##   align the start of buffers to this modulus (default 16384)\n\
    -O ##   start buffers at this offset from the modulus (default 0)\n\
    -p ##   port number to send to or listen at (default 5001)\n\
    -b ##   set socket buffer size (if supported)\n\
    -F      loop for ever\n\
    -D      set TCP_NODELAY socket option\n\
    -s      (Ptcp6 -s) : server mode (mirror) \n\
    -c      (Ptcp6 -c) : client mode (generator) \n\
    -v      verbose: print more statistics\n\
    -n ##   number of iterations\n\
    -L      create and append all results to a file named ttcp_log \n\
";

fprintf(stdout,Usage);
}
/*---------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/
void
open_log()
{static long sysTicks;
 sprintf(logfile,"%s_%s",logfile_head,client?"c":"s");

 fprintf(stdout,"open the log file >%s<\n",logfile);

 if ((fplog = fopen(logfile,"r")) == NULL)
 
   {if ((fplog = fopen(logfile,"a+")) == NULL)
     {fprintf(stdout,"Failure : creation of the file >%s< \n",logfile );
      exit(1);
     }
    else
     {fprintf(fplog," creation date : ");
     /* get date */
     time(&sysTicks);
     tms = localtime(&sysTicks);
     fprintf(fplog,"%02d-%02d-%02d %02d:%02d\n",
            tms->tm_mday, tms->tm_mon, tms->tm_year,tms->tm_hour, tms->tm_min);
     fprintf(fplog,"format\n");
     fprintf(fplog,"buflen, nbuf(byte), bufalign(byte), bufoffset(byte)\n");
     fprintf(fplog,"port, sockbufsize(byte), validity, round(usec), ");
     fprintf(fplog,"latency(usec), rate Mbytes/sec, ");
     fprintf(fplog,"hours*3600+min*60+sec\n\n");
     }
  } /* file already exist */
   else 
   {fclose (fplog);
     if ((fplog = fopen(logfile,"a+")) == NULL)
       {fprintf(stdout,"Failure : access of the file >%s< \n",logfile );
        exit(1);
           }
   }
} 
/*-------------------------------------------------------------------------*/
/* unix like commands under WindowsNT -------------------------------------*/

#if defined(__NTVIS__)

/*-------------------------------------------------------------------------*/
static void
error(char *pch)
{
    if (!opterr) {
        return;        // without printing
    }
    fprintf(stderr, "%s: %s: %c\n",
        (NULL != progname) ? progname : "getopt", pch, optopt);
}

/*-------------------------------------------------------------------------*/
int
getopt(int argc, char **argv, char *ostr)
{
    static char *place = EMSG;    /* option letter processing */
    register char *oli;            /* option letter list index */

    if (!*place) {
        // update scanning pointer
        if (optind >= argc || *(place = argv[optind]) != '-' || !*++place) {
            return EOF; 
        }
        if (*place == '-') {
            // found "--"
            ++optind;
            return EOF;
        }
    }

    /* option letter okay? */
    if ((optopt = (int)*place++) == (int)':'
        || !(oli = strchr(ostr, optopt))) {
        if (!*place) {
            ++optind;
        }
        error("illegal option");
        return BADCH;
    }
    if (*++oli != ':') {    
        /* don't need argument */
        optarg = NULL;
        if (!*place)
            ++optind;
    } else {
        /* need an argument */
        if (*place) {
            optarg = place;        /* no white space */
        } else  if (argc <= ++optind) {
            /* no arg */
            place = EMSG;
            error("option requires an argument");
            return BADCH;
        } else {
            optarg = argv[optind];        /* white space */
        }
        place = EMSG;
        ++optind;
    }
    return optopt;            // return option letter
}
/*--------------------------------------------------------------------------*/
int
gettimeofday(tp, zp)
struct timeval *tp;
struct timezone *zp;
{
  struct _timeb timeptr;
 
  _ftime(&timeptr);
  tp->tv_sec = timeptr.time;
  tp->tv_usec = timeptr.millitm * 1000;
  return(1);
}

#endif /* __NTVIS__ */

/*--------------------------------------------------------------------------*/
